/* -*-C++-*- */
/*
 *  Copyright (c) 1999 Regents of the University of California.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgment:
 *       This product includes software developed by the Imaging and
 *       Distributed Collaboratories Group at Lawrence Berkeley
 *       National Laboratory.
 *  4. Neither the name of the University nor of the Laboratory may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *
 */


/*
* A class to store the unique ID's and secrets, and to answer
* access questions. A client receives access if the hash of 
* it's command concatenated with the secret is equal to the hash
* sent in the packet, and the entry has not timed out.
* author: Keith Jackson
* version: 1.0
* date: 9/2/98
*/

#ifndef SecurityInfo_h
#define SecurityInfo_h

#include <map>
#include "openssl/evp.h"

class SecurityInfo {

public:
 
/*
* Destructor: free's the memory in the hash table. 
* This memory was allocated in the setData() method.
* Note: Only the values need to be freed.  the key's
* are const pointers.
*/

  ~SecurityInfo();

/*
* A method to set the data into the hash table.  The SHA-1
* hash of the DN is used as the key, and the masterSecret
* is the value.  
* DN - The Distinguished Name of the client.  It is hashed
* to become the unique ID used as the key in the hash table.
* masterSecret - This is the secret constructed during the
* SSL handshake.  The secret is associated with the ID
* in the hash table.
* The memory for the buckets in the hash table is allocated
* in this method.  It is free'd in the destructor.
* returns: Returns 1 if able to set the data correctly, and
* 0 otherwise.
*/
  
  int setData(const char* DN, const char* masterSecret);

/*
* A method to make access decisions based on the hash table 
* constructed during the SSL handshakes.  First it parses
* secData, to seperate the ID from the hash.  It then checks
* to see that there is an entry in the hash table for the ID.
* If so it returns the secret.  It not the method returns false,
* and access is denied.  Next the master secret is appended to 
* command, and a SHA-1 hash is taken of this string.  This hash is
* compared to the hash passed in the packet.  If they match, access 
* is granted.  If not access is denied.
* secData - The ID and the hash value from the packet.  The
* format is ID:*:hash.
* command - The full packet minus secData.  This is the data
* that will have the secret appended to it, and be hashed upon.
* returns: true, it access is allowed, false otherwise.
*/

  bool checkAccess(const char* secData, const char* command);

private:

// A hash table to hold the data.
  std::map<const char*, char*> htable;

// A message digest context object.
  EVP_MD_CTX ctx;
};
#endif
