/* -*-C++-*- */
/*
 *  Copyright (c) 1999 Regents of the University of California.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgment:
 *       This product includes software developed by the Imaging and
 *       Distributed Collaboratories Group at Lawrence Berkeley
 *       National Laboratory.
 *  4. Neither the name of the University nor of the Laboratory may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *
 */

/*
 *  This TestInfo class handles the variables
 *  that are common to all network tests. 
 */
#include "TestInfo.h"
#include "Results.h"

#include <iostream>
#include <cstring>

using std::cerr;
using std::endl;


/* -------------------------------------------------------------------
 *  TestInfo: Base Class for IperfTestInfo and PingTestInfo
 *
 *
 * ------------------------------------------------------------------- */

TestInfo::TestInfo( void )
    : MyDebug("TestInfo: "), mBufLenSet(false), mIsSender(false), mLocalName(string("")), 
	mRemoteTestPeer(string("")), mLocalTestInterface( string("")),
	mTest_Params( string("")),delay_secs(0), mPort(3642), 
	mBufferLen(1024), mReportInterval(20.0), mTermination(Timed), mTestAmount(10), 
	mDebugLevel(0), rContainer(NULL)
{
}

TestInfo::TestInfo(bool val, const string& address, const string& interface, int port,
		   int bufferlen, double reportInterval, TerminationTypes termVal, 
		   double testAmount)
    : MyDebug("TestInfo: "), mBufLenSet(false), mIsSender(val), mLocalName(string("")),
	mRemoteTestPeer(address), mLocalTestInterface(interface),
	mTest_Params( string("")), delay_secs(0), mPort(port), 
	mBufferLen(bufferlen), mReportInterval(reportInterval), mTermination(termVal), 
	mTestAmount(testAmount), rContainer(NULL)
{
}

TestInfo::TestInfo( const TestInfo& tInfo ): MyDebug("TestInfo: "), mDebugLevel(0)
{
    printDebug(4, "Copy Constructor \n");

    mBufLenSet      	= tInfo.mBufLenSet;
    mIsSender       	= tInfo.mIsSender;

    mLocalName	    	= tInfo.mLocalName;
    mRemoteTestPeer 	= tInfo.mRemoteTestPeer;
    mLocalTestInterface = tInfo.mLocalTestInterface;
    mTest_Params     	= tInfo.mTest_Params;

    delay_secs	    	= tInfo.delay_secs;
    mPort           	= tInfo.mPort;
    mBufferLen      	= tInfo.mBufferLen;
    mReportInterval 	= tInfo.mReportInterval;
    mTermination    	= tInfo.mTermination;
    mTestAmount     	= tInfo.mTestAmount;
    rContainer      	= tInfo.rContainer;
}

TestInfo::~TestInfo(void)
{
   rContainer = NULL;
}

/* -------------------------------------------------------------------
 *  printVariables
    bool mBufLenSet;
    bool mIsSender; // boolean indication of whether this is the sender

    string mLocalName;
    string mRemoteTestPeer;      // address/name of remote test host
    string mLocalTestInterface;  // which local interface to bind to for the test

    int              delay_secs;      // seconds to wait before test begins
    int              mPort;           // port to use for test
    int              mBufferLen;      // size of each message sent
    double           mReportInterval; // frequency for reporting test results
    TerminationTypes mTermination;    // termination mode
    double           mTestAmount;     // # of bytes, # of messages, or time for test to run
    int              mDebugLevel;     // how much debugging info to print out during test
    Results*         rContainer;
 *
 * ------------------------------------------------------------------- */

ostream &
TestInfo::printVariables(ostream & stream)
{
   stream << "mBufLenSet?          = " << ((mBufLenSet)? "Yes" : "No") << endl;
   stream << "Sender?              = " << ((mIsSender)? "Yes" : "No") << endl;
   stream << "Local Name           = " << mLocalName << endl;
   stream << "Remote TestPeer      = " << mRemoteTestPeer << endl;
   stream << "Local Test Interface = " << mLocalTestInterface << endl;
   stream << "Port                 = " << mPort << endl;
   stream << "BufferLen            = " << mBufferLen << endl;
   stream << "ReportInterval       = " << mReportInterval << endl;
   stream << "Termination          = " << mTermination << endl;
   stream << "TestAmount           = " << mTestAmount << endl;
   stream << "DebugLevel           = " << mDebugLevel << endl;
   return stream;
}
/* -------------------------------------------------------------------
 *  Overloaded Assignment Operator
 *
 * ------------------------------------------------------------------- */

TestInfo&
TestInfo::operator=( const TestInfo& tInfo )
{
    printDebug(4, "Overloaded =\n");

    mBufLenSet      = tInfo.mBufLenSet;
    mIsSender       = tInfo.mIsSender;

    mLocalName	    = tInfo.mLocalName;
    mRemoteTestPeer = tInfo.mRemoteTestPeer;
    mLocalTestInterface = tInfo.mLocalTestInterface;
    mTest_Params   = tInfo.mTest_Params;

    delay_secs	    = tInfo.delay_secs;
    mPort           = tInfo.mPort;
    mBufferLen      = tInfo.mBufferLen;
    mReportInterval = tInfo.mReportInterval;
    mTermination    = tInfo.mTermination;
    mTestAmount    = tInfo.mTestAmount;
    mDebugLevel    = tInfo.mDebugLevel;
    rContainer     = tInfo.rContainer;

    return *this;
}

/* -------------------------------------------------------------------
 *  setUserParams:  mTest_Params must be set before entering.
 *
 * ------------------------------------------------------------------- */

void
TestInfo::setUserParams( void )
{
    string tmpParam( mTest_Params );
    string tokenDelims( "-" );
    string tokenNum( "0123456789" );
    string number;
    string::size_type begIdx;

    printDebug(4, "setUserParams params: %s\n", tmpParam.c_str());

    begIdx = tmpParam.find_first_of( tokenDelims, 0 );
    if( begIdx == string::npos )
        return;

    while( true )
    {
	begIdx = tmpParam.find_first_not_of( tokenDelims, begIdx );
	if( begIdx == string::npos )
	    return;

        printDebug(2, "setUserParams option: \n");

        switch ( tmpParam[begIdx] )
        {
            // CLIENT & SERVER OPTIONS *****************************
            case 'd': // delay in seconds
		getToken( number, tmpParam, tokenNum, begIdx );
                delay_secs = atoi( number.c_str() );
                break;
            default:
                cerr<<"Error Incorrect parameter: "<< endl;
                break;
        }

        begIdx = tmpParam.find_first_of( tokenDelims, begIdx );
        if( begIdx == string::npos )
            break;
    }

    return;
}

/* -------------------------------------------------------------------
 *  getToken: take as input params containing a set of tokens,
 *            a set of symbols to search for, and a beginning index.
 *            It will search for a match and return a token containing
 *            all characters continuous from the first match.
 *
 * ------------------------------------------------------------------- */

void
TestInfo::getToken( string& ans, const string& params, const string& symbols,
                         const string::size_type& begIdx )
{
    string::size_type frontIdx = params.find_first_of( symbols, begIdx );

    if( frontIdx == string::npos )
    {
        ans = "";
        return;
    }

    string::size_type  rearIdx = params.find_first_of( ' ', frontIdx );

    if( rearIdx == string::npos )
        rearIdx = params.length();

    ans.assign( params, frontIdx, (rearIdx-frontIdx) );

    return;
}
