/* -*-C++-*- */
/*
 *  Copyright (c) 1999 Regents of the University of California.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgment:
 *       This product includes software developed by the Imaging and
 *       Distributed Collaboratories Group at Lawrence Berkeley 
 *       National Laboratory.
 *  4. Neither the name of the University nor of the Laboratory may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS;OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *
 *  $Id: udpnet.cc,v 1.8 2001/04/19 18:36:24 boverhof Exp $
 */


#include "udpnet.h"
#include <string>
#include <iostream>
#include <cstring>
#include <cstdio>
#include <cstdlib>

using std::cerr;
using std::endl;
using std::string;
using std::memset;
using std::memcpy;
using std::exit;
using std::printf;

#define IN_CLASS_D(i)            (((long)(i) & 0xf0000000) == 0xe0000000)


UDPNet::UDPNet(int fd, int timeout, short p) 
    : Network(fd,timeout,p), mTtl(127), mMulticast(false)
{
    setClassName( "UDPNet: " );
}
    
UDPNet::~UDPNet()
{
}

int UDPNet::openSocket()
{
    return openSocket( (const char *)NULL );
}

// This routine opens a Datagram socket; can be used for udp or ipmulticast. 
// If an empty string is passed in this socket can only be used for
// udp receives.  If a multicast address is supplied then the address
// will be joined.  All messages sent on the socket will be sent to
// the address supplied. 

int UDPNet::openSocket(const string& address )
{
    struct hostent *hp;
    struct sockaddr_in sock_sa;
    unsigned long lngaddr;
    char ssval;
    long on=1;
    int err;
    struct ip_mreq mreq;

    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock < 0)
    {
        printError("UDPNet::openSenderSock--socket() failed");
	return (-1);
    }

    if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char *)&on,(int)sizeof(on)) < 0) 
    {
        printError("UDPNet--setsockopt failed for SO_REUSEADDR");
	return (-1);
    }
    
    memset(&peer, 0, sizeof(peer));
    peer.sin_family = AF_INET;
    peer.sin_port = htons(port);

    if( address != "" )
      {
	if( isdigit( address[0] ) )
	  {
	    // This address is specified in dotted decimal notation 
	    lngaddr = (unsigned long)inet_addr(address.c_str());
	    // first check to see if this is a multicast address
	    if (IN_CLASS_D(ntohl(lngaddr)))
	      {
		mMulticast = true;
		peer.sin_addr.s_addr = htonl(INADDR_ANY);
		// set the ttl for the session
		ssval = mTtl;
		err = setsockopt(sock,IPPROTO_IP, IP_MULTICAST_TTL,
				  (char *)&ssval, sizeof(ssval));
		if ( err < 0)
		  {
		      cerr << "setsockopt error setting ttl" << endl;
		      exit( -1 );
		  }
		// turn off loopback
		ssval = 0;
		if (setsockopt(sock,IPPROTO_IP, IP_MULTICAST_LOOP,
			       (char *)&ssval, sizeof(ssval)) == -1 )
		  {
		      cerr << "bcast setsockopt2 error" << endl;
		      exit( -1 );
		  }
		// join the multicast group address
		mreq.imr_multiaddr.s_addr = lngaddr;
		mreq.imr_interface.s_addr = INADDR_ANY;
		if (setsockopt(sock,IPPROTO_IP, IP_ADD_MEMBERSHIP,
			       (char *)&mreq, sizeof(struct ip_mreq)) == -1 )
		  {
		      cerr << "bcast setsockopt3 error" << endl;
		      exit( -1 );
		  }
		
	      }
	    else // address of a machine specified by digits (not multicast)
	      peer.sin_addr.s_addr = htonl( lngaddr );
	  }
	else
	  {
	    // This address specifies a machine by name so we need
	    // to translate it 
	    hp = gethostbyname( address.c_str() );
	    if (!hp) 
	        {
		printError("UDPNet--gethostbyname() failed");
		return (-1); 
		}
	    memcpy(&peer.sin_addr, hp->h_addr, hp->h_length);
	  }
      }
    else // no address spcified in call to open socket
      {
	peer.sin_addr.s_addr = htonl(INADDR_ANY);
      }

    if( address != "" & !mMulticast)
      {
	memset(&sock_sa, 0, sizeof(sock_sa));
	sock_sa.sin_family = AF_INET;
	sock_sa.sin_addr.s_addr = htonl(INADDR_ANY);
	sock_sa.sin_port = htons( port );
	if (bind(sock, (struct sockaddr *)&sock_sa, sizeof(sock_sa)) < 0)
	    {
	    printError("UDPNet::openSenderSock--bind() failed");
	    return (-1);
	    }
      }
    else
      {
	if (bind(sock, (struct sockaddr *)&peer, sizeof(peer)) < 0)
	    {
	    printError("UDPNet--bind failed");
	    return (-1);
	    }
	if( mMulticast )
	    peer.sin_addr.s_addr = lngaddr;
      }

    return 0; 	
}
     	
// Routine to close the socket if it was not already closed
void UDPNet::closeSocket()
{
    if( sock > 0 )
        close( sock );
    sock = -1;
}
