/* -*-C++-*- */
/*
 *  Copyright (c) 1999 Regents of the University of California.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgment:
 *       This product includes software developed by the Imaging and
 *       Distributed Collaboratories Group at Lawrence Berkeley 
 *       National Laboratory.
 *  4. Neither the name of the University nor of the Laboratory may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *
 *  $Id: TestModule.cc,v 1.20 2001/09/05 23:35:33 boverhof Exp $
 */

#include "TestModule.h"
#include <string>
#include <iostream>
#include <wait.h>
#include "Display_funcs.h"
#include "Results.h"
#include "funcs.h"

//ping 
#include "PingTestInfo.h"
#include "udpping.h"
#include "multiping.h"

//iperf
#include "IperfTestInfo.h"
#include "Listener.hpp"
#include "Client.hpp"
#include "Server.hpp"

// defined in Display_funcs.h
extern const string REPORTS;
extern const string MASTER;
extern const string IPERF;
extern const string UDPPING;
extern const string MULTIPING;
extern const string TUNEDTCP;
extern const string TRACEROUTE;
extern const string PIPECHAR;

extern const string IPERF_PORT;
extern const string UDPPING_PORT;
extern const string EMPTY;
extern const string FAILED;
extern const string FINISHED;
extern const string DELIM;

using std::cout;
using std::cerr;
using std::endl;


TestModule::TestModule(void): MyDebug("TestModule"), MyThread(), c_info_ptr(NULL)
{
    pthread_mutex_init( &data_mutex, NULL );
    pthread_mutex_init( &listener_mutex, NULL );
}
TestModule::~TestModule(void)
{
    pthread_mutex_destroy( &data_mutex );
    pthread_mutex_destroy( &listener_mutex );
}


/* -----------------------------------------------------
 * startTest: Called by main, with CoordInfo filled out
 *	lock data_mutex
 *	Begin() will launch a thread that calls ThreadMain
 *        -> MyThread::StartThread -> pthread_create -> ThreadMain
 * ----------------------------------------------------- */

void
TestModule::startTest( CoordInfo coordinfo )
{ 
    printDebug(2, "Start Test\n");
    pthread_mutex_lock( &data_mutex );
    if( c_info_ptr )
	delete c_info_ptr;
    c_info_ptr = new CoordInfo( coordinfo );
    Begin();
    //pthread_mutex_unlock( &data_mutex );
    printDebug(2, "Out Start Test\n");
}


/* -----------------------------------------------------
 * ThreadMain: Thread routine
 *    Create local copy of CoordInfo, release data_mutex, and
 *    pass it by ref to other methods.
 *
 * ----------------------------------------------------- */

void
TestModule::ThreadMain()
{
    printDebug(2, "ThreadMain\n");
    //pthread_mutex_lock( &data_mutex );
    CoordInfo c_info( *c_info_ptr );
    pthread_mutex_unlock( &data_mutex );
    c_info.printVariables();
    doTest( c_info );
    printDebug(2, "ThreadMain done doTest/n");
    return;
}

/* ---------------------------------------------------------
 *  Cleanup - the intent of this routine is to be called when a thread
 *  is stopped prematurely by the user.  It prints the final statistics 
 *  of the ping test before the thread get destroyed.  This routine is
 *  currently not working.
 * ---------------------------------------------------------*/

void 
TestModule::Cleanup( void )
{
    printDebug(2, "Cleanup\n");
    MyThread::Cleanup();
    return;
}

void
TestModule::doTest( CoordInfo& coordinfo )
{
    printDebug(2, "doTest\n");
    string test  = coordinfo.getTestName();

    if( test == IPERF )
        doTest_Iperf( coordinfo );
    else if( test == UDPPING )
        doTest_Udpping( coordinfo );
    else if( test ==  MULTIPING )
        doTest_Multiping( coordinfo );
    else if( test == TUNEDTCP )
        doTest_Tuned( coordinfo );
#ifdef PATH_TRACEROUTE
    else if( test == TRACEROUTE )
	doTest_Traceroute( coordinfo );
#endif
#ifdef PATH_PIPECHAR
    else if( test == PIPECHAR )
	doTest_Pipechar( coordinfo );
#endif
    else
	printDebug(2, "ERROR doTest no such test\n");
}


/* -----------------------------------------------------------------
 * doTest_Iperf: Create & Fill in testinfo object of test before launch test
 *
 *  iperfinf->getClientThreads() - never examined by Listener??
 * ----------------------------------------------------------------- */

int
TestModule::doTest_Iperf( CoordInfo& coordinfo )
{
	static Listener *theListener(NULL);

        IperfTestInfo *iperfinf = new IperfTestInfo();
        iperfinf->setGroupID( getThreadID() );

        iperfinf->setLocalName( coordinfo.getLocalHost() );
        iperfinf->setResults( coordinfo.getResults_p() );
        iperfinf->setPort( atoi(IPERF_PORT.c_str()) );

        string param_str = coordinfo.getIperfParams();
        if(  param_str != EMPTY ){
            iperfinf->setTestParams( param_str );
            iperfinf->setUserParams();
        }
        // Screw up Listener??
        Client::SetNumThreads( iperfinf->getClientThreads());

        printDebug(2, "Starting an iperf test\n");
#ifdef WIN32
        WSADATA wsaData;
        int rc = WSAStartup( 0x202, &wsaData );
        FAIL_errno( rc == SOCKET_ERROR, "WSAStartup" );
        SetConsoleCtrlHandler( sig_dispatcher, true );
#endif
        // perform any cleanup when quitting Iperf ??Windows Only
        //atexit( cleanup );

        if( coordinfo.getIsTestSender() )
        {
            iperfinf->setRemoteTestPeer( coordinfo.getTestReceiver() );
            iperfinf->setIsSender( true );

	// ->  MUTEX?
            pthread_mutex_lock( &listener_mutex );
            // start up a listener
	    if( theListener==NULL ){
                printDebug(2, "theListener == NULL\n");
		theListener = new Listener( iperfinf  );
	    }
	    else
                printDebug(2, "theListener is Already alive\n");

            pthread_mutex_unlock( &listener_mutex );
	// <-  MUTEX?
	    
	    theListener->Start();

#ifdef HAVE_THREAD
            if ( Thread::NumUserThreads() > 0 ){
                printDebug(2, "Waiting for server threads to complete.\n" );
                fflush( 0 );
            }
#endif
	    // wait for other (client, server) threads to complete
	    Thread::Joinall();
	    printDebug(2, "after Joinall\n");

	// ->  MUTEX?
            pthread_mutex_lock( &listener_mutex );
	    if( theListener != NULL ){
		delete theListener;
		theListener = NULL;
		printDebug(2, "Delete theListener %d\n");
	    }
	    else
		printDebug(2, "theListener is Dead%d\n");
            pthread_mutex_unlock( &listener_mutex );
	// <-  MUTEX?
        }
        else if( coordinfo.getIsTestReceiver() )
        {
            iperfinf->setRemoteTestPeer( coordinfo.getTestSender() );
            iperfinf->setIsSender( false );

	    //iperfinf->printVariables(cout);
            // Ask User for test parameters
            printDebug(2, "\tstart up Client\n");
            printDebug(2, "\tnumber of threads %d\n", iperfinf->getClientThreads());

	    // LOCK: protect Client::sNumThreads
            Client::SetNumThreads( iperfinf->getClientThreads());
            sleep(1);
            for( int i = 0; i < iperfinf->getClientThreads(); i++ ){
                // start up a client
                Client* theClient = new Client( iperfinf );
                theClient->DeleteSelfAfterRun();
                theClient->Start();
            }
	    // UNLOCK

	    printDebug(2, "Client before Joinall\n");
	    // wait for other (client, server) threads to complete
	    Thread::Joinall();
	    printDebug(2, "Client after Joinall\n");
        }
        else
            printDebug(2,"neither server nor client mode was specified");

        //  End-of-Test Report: Accumulation of threads
        if( (iperfinf->getResults())->getIsReportEndOfTest() )
            (iperfinf->getResults())->iperfResults_EndOfTest( iperfinf->getGroupID() );

        delete iperfinf;

    return 0;
}


/* ---------------------------------------------------------------------------------
 * doTest_Udpping
 *
 * --------------------------------------------------------------------------------- */

int
TestModule::doTest_Udpping( CoordInfo& coordinfo )
{
        printDebug(2, "UDPPING TEST pinginf\n");
        PingTestInfo pinginf;

        pinginf.setLocalName( coordinfo.getLocalHost() );
        pinginf.setResults( coordinfo.getResults_p() );
        pinginf.setIsMulti( false );
        //pinginf->setPort( UDPPING_PORT );

        printDebug(2, "server??\n");
	if(coordinfo.getIsTestReceiver()){
            pinginf.setRemoteTestPeer( coordinfo.getTestSender() );
	    pinginf.setIsSender( false );
	}
	else if( coordinfo.getIsTestSender() ){
            pinginf.setRemoteTestPeer( coordinfo.getTestReceiver() );
	    pinginf.setIsSender( true );
	}

        printDebug(2, "get Params\n");
	string param_str = coordinfo.getPingParams();
        // Get User Parameters
        if(  !param_str.empty() ){
            pinginf.setTestParams( param_str );
            pinginf.setUserParams();
	}
	else
            cout << "ERROR doTest: udpping no params\n";

        UDPPing pingtest( pinginf );
        printDebug(2, "doTest_udpping Run ping Test\n");
        pingtest.ThreadMain();
	pingtest.printResults();

        //delete pingtest;
        //delete pinginf;

    return 0;
}


/* ---------------------------------------------------------------------------------
 * doTest_Multiping
 *
 * --------------------------------------------------------------------------------- */

int
TestModule::doTest_Multiping( CoordInfo& coordinfo )
{
        printDebug(2, "UDPPING TEST pinginf\n");
        PingTestInfo mpinginf;
        mpinginf.setLocalName( coordinfo.getLocalHost() );
        mpinginf.setResults( coordinfo.getResults_p() );
        mpinginf.setMcastTTL( 20 );
        mpinginf.setIsMulti( true );
        //mpinginf->setPort( UDPPING_PORT );

	if(coordinfo.getIsTestReceiver()){
            mpinginf.setRemoteTestPeer( coordinfo.getTestSender() );
	    mpinginf.setIsSender( false );
	}
	else if( coordinfo.getIsTestSender() ){
            mpinginf.setRemoteTestPeer( coordinfo.getTestReceiver() );
	    mpinginf.setIsSender( true );
	}
	else
	    return -1;

	string param_str = coordinfo.getMultiParams();
        // Get User Parameters
        if(  !param_str.empty() ){
            mpinginf.setTestParams( param_str );
            mpinginf.setUserParams();
	}
	else
            cout << "ERROR doTest: multiping no params\n";

        MULTIPing multitest( mpinginf );
        multitest.ThreadMain();
	multitest.printResults();

	//mpinginf->printVariables(cout);
        //MULTIPing *multitest = new MULTIPing( mpinginf );
        //multitest->Begin();
        //multitest->endTest();
        //delete multitest;
        //delete mpinginf;

    return 0;
}


/* ---------------------------------------------------------------------------------
 * doTest_Tuned
 *
 * --------------------------------------------------------------------------------- */

int
TestModule::doTest_Tuned( CoordInfo& coordinfo )
{
    printDebug(2, "TUNED TEST pinginf\n");

    // ************** Get RTT ****************************
    PingTestInfo pinginf;
    pinginf.setLocalName( coordinfo.getLocalHost() );
    pinginf.setResults( coordinfo.getResults_p() );
    pinginf.setIsMulti( false );
    //pinginf->setPort( UDPPING_PORT );

    if(coordinfo.getIsTestReceiver()){
        pinginf.setRemoteTestPeer( coordinfo.getTestSender() );
        pinginf.setIsSender( false );
    }
    else if( coordinfo.getIsTestSender() ){
        pinginf.setRemoteTestPeer( coordinfo.getTestReceiver() );
        pinginf.setIsSender( true );
    }
    else
    	return -1;

    string param_str = coordinfo.getTunedParams();
    printDebug(2, "TunedParams: %s\n", param_str.c_str());
    string tparams = getToken( param_str, DELIM, 0);

    // Get User Parameters
    if(  !tparams.empty() ){
        pinginf.setTestParams( tparams );
        pinginf.setUserParams();
    }

    //pinginf->printVariables(cout);
    //UDPPing *pingtest = new UDPPing( pinginf );
    //pingtest->Begin();
    //pingtest->endTest();

    UDPPing pingtest( pinginf );

    // ************** Tuned Iperf ****************************
    IperfTestInfo *iperfinf = new IperfTestInfo();
    iperfinf->setLocalName( coordinfo.getLocalHost() );
    iperfinf->setResults( coordinfo.getResults_p() );
    iperfinf->setPort( atoi(IPERF_PORT.c_str()) );
    tparams = getToken( param_str, DELIM, 1);

    // Get User Parameters
    if(  param_str != EMPTY ){
        iperfinf->setTestParams( param_str );
        iperfinf->setUserParams();
    }

    printDebug(2, "Starting TCP iperf test\n");
    double targetbw    = atof( (getToken(param_str, DELIM, 2)).c_str() );
    double avertt      = pingtest.getAverageRTT();
    int    desired_win = static_cast<int>(2 * targetbw *1024 * avertt);

    printDebug(2, "Target bandwidth is    : %f.1 Mb/sec\n", targetbw);
    printDebug(2, "Average RTT            : %f.1 Mb/sec\n", avertt);
    printDebug(2, "Setting window size to : %d KBytes\n", desired_win);

    if( desired_win < 512 )
        desired_win = 512;

    iperfinf->setTCPWindowSize( desired_win );
    Client::SetNumThreads( iperfinf->getClientThreads());
    printDebug(2, "main: Starting an iperf test\n");
    Listener *theListener = NULL;

    // signal handlers quietly exit on ^C and kill
    // these are usually remapped later by the client or server
    //my_signal( SIGTERM, sig_exit );
    //my_signal( SIGINT,  sig_exit );

#ifdef WIN32
    WSADATA wsaData;
    int rc = WSAStartup( 0x202, &wsaData );
    FAIL_errno( rc == SOCKET_ERROR, "WSAStartup" );
    SetConsoleCtrlHandler( sig_dispatcher, true );
#endif
    // perform any cleanup when quitting Iperf ?Windows Only
    //atexit( cleanup );

    if( coordinfo.getIsTestSender() ){
        iperfinf->setRemoteTestPeer( coordinfo.getTestReceiver() );
        iperfinf->setIsSender( true );

        // start up a listener
        theListener = new Listener( iperfinf  );
        theListener->Start();

#ifdef HAVE_THREAD
        if ( Thread::NumUserThreads() > 0 ){
            fflush( 0 );
	}
#endif
    }
    else{
        iperfinf->setRemoteTestPeer( coordinfo.getTestSender() );
        iperfinf->setIsSender( false );

        // Ask User for test parameters
        Client::SetNumThreads( iperfinf->getClientThreads());

        for( int i = 0; i < iperfinf->getClientThreads(); i++ ){
            // start up a client
            sleep(1); // a hack for listener to be waiting on accept
            Client* theClient = new Client( iperfinf );
            theClient->DeleteSelfAfterRun();
            theClient->Start();
    	}
    }

    // wait for other (client, server) threads to complete
    Thread::Joinall();
    DELETE_PTR( theListener );

    delete iperfinf;
    //delete pingtest;
    //delete pinginf;

    return 0;
}


/* ---------------------------------------------------------------------------------
 * doTest_Traceroute:
 *
 * --------------------------------------------------------------------------------- */

#ifdef PATH_TRACEROUTE
void
TestModule::doTest_Traceroute( CoordInfo& coordinfo )
{
    string path( PATH_TRACEROUTE );        
    int fd[2];

    if( pipe(fd) < 0 ){
	    cout << "ERROR: pipe failed\n";
	    return;
 	}

	int pid = fork();
	if( pid == 0 ){ // child
	    printDebug(2,"Child process\n");
	    close(fd[0]);
	    dup2(fd[1], 1);
	    dup2(fd[1], 2);

            //Take Parameters -> Turn into arg list
	    char* args[30];
	    args[0] = (char*) malloc( path.length()+1 );
	    strcpy( args[0], path.c_str());

	    string temp( coordinfo.getTestReceiver() );
            args[1] = (char*) malloc( temp.length()+1 );
	    strcpy( args[1], temp.c_str());

	    string params(coordinfo.getTraceParams());
	    string::size_type begIdx(0), endIdx(0);
	    (params.empty())? begIdx=string::npos:
		    begIdx = params.find_first_not_of(' ', begIdx);
	    int i;
	    for( i=2; i<30 && endIdx!=string::npos && begIdx!=string::npos; i++)
	    {
		endIdx = params.find_first_of(' ', begIdx);
		(endIdx==string::npos)? endIdx=params.length(): endIdx;
		temp.assign( params, begIdx, (endIdx-begIdx) );
		args[i] = (char*) malloc( temp.length()+1 );
		strcpy( args[i], temp.c_str());
		begIdx = params.find_first_not_of(' ', endIdx);
	    }
	    args[i] = NULL;
	    cout << "path: " << path << endl;
	    int err = execv( path.c_str(), args ); 
	    cout << "**ERROR(" << err << "): Test Failed" << endl;
	    exit(-1);
	}
	else if( pid < 0 ){ // error
	    cout << "ERROR (" << pid << "): forking process failed\n";
	}
	else{ // parent
 	    int x;
	    string hold("");
	    char buf[100];

	    printDebug(2, "Parent process\n");
	    close(fd[1]);
            memset( buf, 0, sizeof(buf) );
            while( (x=read(fd[0], buf, sizeof(buf))) > 0 ){
                hold.append(buf);
                memset( buf, 0, sizeof(buf) );
            }
            int status; 
	    wait( &status );
	    
	    if( hold.find( "**ERROR") != string::npos )
		cout << "**Traceroute Test Failed" << endl;
            else
	    	(coordinfo.getResults_p())->tracerouteResults
			( false,coordinfo.getLocalHost(),coordinfo.getTraceParams(),hold );
	}
}
#endif


/* ---------------------------------------------------------------------------------
 *  doTest_Pipechar:
 *
 * --------------------------------------------------------------------------------- */

#ifdef PATH_PIPECHAR
void
TestModule::doTest_Pipechar( CoordInfo& coordinfo )
{
    string path( PATH_PIPECHAR );
    int fd[2];
    int i;

    //Take Parameters -> Turn into arg list
    char* args[30];
    args[0] = (char*) malloc( path.length()+1 );
    strcpy( args[0], path.c_str());

    string temp("");

    string params(coordinfo.getTraceParams());
    string::size_type begIdx(0), endIdx(0);
    (params.empty())? begIdx=string::npos: begIdx = params.find_first_not_of(' ', begIdx);

    for( i=1; i<30 && endIdx!=string::npos && begIdx!=string::npos; i++)
    {
        endIdx = params.find_first_of(' ', begIdx);
        (endIdx==string::npos)? endIdx=params.length(): endIdx;
        temp.assign( params, begIdx, (endIdx-begIdx) );
        args[i] = (char*) malloc( temp.length()+1 );
        strcpy( args[i], temp.c_str());
        begIdx = params.find_first_not_of(' ', endIdx);
    }
    args[i] = NULL;

    //  Debug
#if 0
    cout << "Run program: path=" << path << endl;
    for ( int i=0; args[i] != NULL; i++ )
    {
               cout << "args[" << i << "]=" << args[i] << endl;
     }
    cout << flush; 
#endif 
    /* setup the pipe for the parent and child to communicate */
    if( pipe(fd) < 0 ){
            cout << "ERROR: pipe failed\n";
            return;
    }

    int pid = fork();
    if( pid == 0 ){ // child
            printDebug(2,"Child process\n");
            close(fd[0]);
            // redirect stderr, stdout throught the pipe back to the parent
            dup2(fd[1], 1);
            dup2(fd[1], 2);

            // Run the program
            int err = execv( path.c_str(), args );

            cout << "**ERROR(" << err << "): Test Failed" << endl;
            exit(-1);
     } 
     else if( pid > 0 )
     { 
               int status;
               string hold("");
               char buf[100];

               printDebug(2, "Parent process\n");
               close(fd[1]);
               memset( buf, 0, sizeof(buf) );
               // stash the args in the hold buffer
               // hold.append(params);
               while( read(fd[0],buf,sizeof(buf)-1) > 0 )
               {
                   hold.append(buf);
                   memset( buf, 0, sizeof(buf) );
               }
               cout << "(pid=" << pid << ") pipechar output-->" << hold << "<--\n";
               wait( &status );

               if( hold.find( "**ERROR") != string::npos )
                     cout << "**Pipechar Test Failed" << endl;
                else
                      (coordinfo.getResults_p())->pipecharResults( false ,
					coordinfo,
					hold );
        }
	else
            cout << "ERROR (" << pid << "): forking process failed\n";
}

#endif /* PATH_PIPECHAR */
