/* -*-C++-*- */
/*
 *  Copyright (c) 1999 Regents of the University of California.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgment:
 *       This product includes software developed by the Imaging and
 *       Distributed Collaboratories Group at Lawrence Berkeley
 *       National Laboratory.
 *  4. Neither the name of the University nor of the Laboratory may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *
 */

/* ---------------------------------------------------------
 *   MULTIPing:  
 *	Class defining a ip multicast ping test
 *
 * --------------------------------------------------------- */

#include <iostream>
#include <string>
#include "multiping.h"
#include "Results.h"
#include "PingResults.h"
#include "timefuncs.h"

using std::cout;
using std::cerr;
using std::endl;

#define MAXPINGMSG     1024

/* ---------------------------------------------------------
 *  MULTIPing constructor
 *  arguments
 *           serv - boolean indicator of whether this is the test server
 *           address - address to send the pings to and expect them from
 *           port - port to use for the ping test
 *           delay - msec delay between pings
 *
 * --------------------------------------------------------- */

MULTIPing::MULTIPing( PingTestInfo testinf ): MyDebug("MULTIPing"), mPinginf(testinf)
{
    // default number of ping tests (infinite)
    mTestlimit = static_cast<int>( mPinginf.getTestAmount() );

    // how many msecs to delay between each ping test
    // minimum delay will be the round trip time
    mDelaypings = static_cast<int>(1000/mPinginf.getSendRate());

    for( int i=0; i < AVGWINDOW; i++)
      mLastrtt[i] = 0.0;

    // how many ping tests have been run
    mNumbertests = 0;

    // init the ping  multicast sockets
    mUdpcon_send = IPNet( 0, mPinginf.getMcastAddr(), mDelaypings, mPinginf.getPort() );
    mUdpcon_recv = IPNet( 0, mPinginf.getMcastAddr(), mDelaypings, mPinginf.getPort() );

    printDebug( 3, "MULTIPing Constructor Done\n" );
}

MULTIPing::~MULTIPing()
{
}


/* ---------------------------------------------------------
 *  sendPingTest - conducts server side of ping test
 *  Conducts a repetitive ping test and *  stores the results.  
 *  This side calculates the RTT.
 *
 *  multicast version 2 hosts: 1 cycle
 *	server - sends message to IP_address
 *		 receives same message
 *		 receives client's message
 *      client - receive server message
 *		 sends its own message
 *		 receives own message
 * ---------------------------------------------------------*/

int MULTIPing::sendPingTest()
{
    printDebug(3, "sendPingTest()\n");

    struct timeval lasttime, curtime;  // time most recent ping was sent and received respectively
    char buf_send[MAXPINGMSG];  // buffer for storing incoming ping message
    char buf_recv[MAXPINGMSG];  // buffer for storing incoming ping message
    int err;
    int cursleeptime;  // measured in msecs, time to sleep before sending next ping
    int temputime;     // temp variable used in calculating time to sleep between pings
    double prevrtt = -1.0;  // last ping round trip time
    int numberfail = 0;  // number of consecutive pings that have not returned

    string sPeeraddr  =  mPinginf.getRemoteTestPeer();
    string sLocalName =  mPinginf.getLocalName();

    mNumbertests = 0;
    err = mUdpcon_send.openSocket();  // open the ping socket
    if( err < 0 ) {
        cerr << "mUdpcon_send.openSocket() failed \n";
        return -1;
    }
    err = mUdpcon_recv.openSocket();  // open the ping socket
    if( err < 0 ) {
        cerr << "mUdpcon_recv.openSocket() failed \n";
        return -1;
    }

    mUdpcon_recv.setBlockTime( 3, 0 );  // set the receive blocking time

    // send ping test
    memset( (void*)&lasttime, 0, sizeof(struct timeval) );
    memset( (void*)&buf_send, 0, sizeof(buf_send) );
    gettimeofday( &lasttime, NULL );

    string format("%s %s %ld %ld %f");
    sprintf( buf_send, format.c_str(), sLocalName.c_str(), 
	 sPeeraddr.c_str(), lasttime.tv_sec,
	 lasttime.tv_usec, prevrtt );

    //** send message to IP address *************************
    do{
    	err = mUdpcon_send.doSend( buf_send, strlen( buf_send ));
    	if( err < 0 ){
	    numberfail++;
	    printDebug( 3, " 1) mUdpcon_send.doSend  failed");
    	}
    	else
	    printDebug( 3, " 1) Sent message: %s\n", buf_send );
    }while( err<0 && numberfail<10 );

    while( mNumbertests< mTestlimit && numberfail<10 ){
	//** receive a message from multicast *******************
	memset( (void*)&buf_recv, 0, sizeof(buf_recv) );
        memset( (void*)&curtime, 0, sizeof(struct timeval) );

	err = mUdpcon_recv.doRecv( buf_recv, MAXPINGMSG );
	gettimeofday( &curtime, NULL );
	printDebug(3, " curtime=%d.%d\n", curtime.tv_sec, curtime.tv_usec );

	if( err > 0 ){
	    if( strcmp(buf_send, buf_recv) == 0 ) // same message
	    	printDebug(3, " 2) received own message- %s\n", buf_recv );
	    else{ 
	    	printDebug(3, " 3) received bounce message: %s\n", buf_recv );

	        prevrtt    = calculateRTT( buf_recv, &curtime );
		if( prevrtt > 0.0 )
		   ++mNumbertests;

		// Delay Next Send
        	cursleeptime = mDelaypings - (curtime.tv_sec - lasttime.tv_sec )*1000;
        	temputime = (curtime.tv_usec - lasttime.tv_usec )/1000;
        	if( temputime > 0 )
            	    cursleeptime -= temputime;

        	if( mNumbertests != 0 && cursleeptime > 0 )
            	    usleep( cursleeptime * 1000 );

    		// SEND ping test
    		memset( (void*)&lasttime, 0, sizeof(struct timeval) );
    		memset( (void*)&buf_send, 0, sizeof(buf_send) );
    		gettimeofday( &lasttime, NULL );

    		sprintf( buf_send, format.c_str(), sLocalName.c_str(),
         		sPeeraddr.c_str(), lasttime.tv_sec,
         		lasttime.tv_usec, prevrtt );

    		//** send message to IP address *************************
 		do{
    		     err = mUdpcon_send.doSend( buf_send, strlen( buf_send ));
        	     if( err < 0 ){
            	    	numberfail++;
            	    	printDebug(3, " 1) mUdpcon_send.doSend  failed");
        	     }
        	     else{
            	    	printDebug(3, " 1) Sent message: %s\n", buf_send );
			numberfail = 0;
		     }
    		}while( err<0 && numberfail<10 );
        	printDebug(3, "  mTestlimit = %d      mNumbertests = %d \n", mTestlimit, mNumbertests );
	    }
	}
	// SEND A NEW ONE THIS IS A FAIL CONDITION
	else{
            numberfail++;
	    printDebug(3, " 1) sendPingTest - doRecv failed   numberfail=%d\n", numberfail );

	    // Delay Next Send
            cursleeptime = mDelaypings - (curtime.tv_sec - lasttime.tv_sec )*1000;
            temputime = (curtime.tv_usec - lasttime.tv_usec )/1000;

            if( temputime > 0 )
                cursleeptime -= temputime;

            if( mNumbertests != 0 && cursleeptime > 0 )
                usleep( cursleeptime * 1000 );

    	    // SEND ping test
    	    memset( (void*)&lasttime, 0, sizeof(struct timeval) );
    	    memset( (void*)&buf_send, 0, sizeof(buf_send) );
    	    gettimeofday( &lasttime, NULL );

    	    sprintf( buf_send, format.c_str(), sLocalName.c_str(),
            	sPeeraddr.c_str(), lasttime.tv_sec,
            	lasttime.tv_usec, prevrtt );

    	    //** send message to IP address *************************
	    do{
       	    	err = mUdpcon_send.doSend( buf_send, strlen( buf_send ));
            	if( err < 0 ){
            	    numberfail++;
            	    printDebug(3, " 1) mUdpcon_send.doSend  failed");
            	}
            	else
            	    printDebug(3, " 1) Sent message: %s\n", buf_send );

   	    }while( err<0 && numberfail<10 );

   	}

	// if the client has not bounced back a ping in 20 seconds
	if( numberfail > (20000 / mDelaypings )){
	    cerr << "Failed to receive a ping from the client in the last 20 seconds, closing connection to client\n";
	    break;
	}
    }

    if(numberfail>=10){
        for( int i=mNumbertests%AVGWINDOW; i>=0; i++ )
            mLastrtt[i] = -1.0;
    }
    return mNumbertests;
}


/* ---------------------------------------------------------
 *  bouncePingTest:
 * 	Conducts client side of ping test
 *  	This routine receives pings in and bounces them back to the
 *  	originator.  Also in the received ping packet is the most
 *  	recent ping results.  
 *
 * --------------------------------------------------------- */

int MULTIPing::bouncePingTest()
{
    printDebug(3, "bouncePingTest()\n");

    struct timeval rcvtime;  // time the ping packet came in
    struct timeval sndtime;  // time when the ping packet is about to be sent
    struct timeval difftime;  // time spent processing the ping packet
    char buf_recv[MAXPINGMSG];    // the ping packet is stored here
    char buf_send[MAXPINGMSG];    // the ping packet is stored here
    char buf2[MAXPINGMSG];   // used to compose difftime to add to ping packet
    int err;
    char *send_sec, *send_usec; // what time the ping packet was sent by the server
    char *serv, *receiver;  // ping packet from and to fields
    char *msg;       // temporary buffer pointer used while parsing ping packet
    char *prev;      // last ping round trip time
    double dblprev;  // last ping round trip time converted to a double

    string bPeeraddr  = mPinginf.getRemoteTestPeer();
    string bLocalName =  mPinginf.getLocalName();

    err = mUdpcon_send.openSocket();  // open socket for ping test
    if( err < 0 ){
	cerr << "mUdpcon_send.openSocket( ) failed \n";
	return -1;
    }

    err = mUdpcon_recv.openSocket();  // open socket for ping test
    if( err < 0 ){
	cerr << "mUdpcon_recv.openSocket() failed \n";
	return -1;
    }

    printDebug(3, "bouncePingTest socket open\n");

    // wait a minimum of 1 second for the ping packet
    // set the receive blocking time
    if( mDelaypings < 1000 )
        mUdpcon_send.setBlockTime( 1, 0 );
    else
        mUdpcon_send.setBlockTime( mDelaypings/1000, ((mDelaypings%1000)*1000) );  

    if( mDelaypings < 1000 )
        mUdpcon_recv.setBlockTime( 1, 0 );
    else
        mUdpcon_recv.setBlockTime( mDelaypings/1000, ((mDelaypings%1000)*1000) );  

    mNumbertests = 0;                
    int numberfail  = 0;  // number of consecutive lost ping packets

    // message    server  multicastIPaddr  send_seconds   send_useconds   previous
    //*********** receive multicast reply to send?? *******************

    string format1("%s %s %s %s");
    string format2(" %ld %ld");
    do{
	memset( (void*)&rcvtime, 0, sizeof(struct timeval) );
        memset( (void*)&buf_recv, 0, sizeof(buf_recv) );
        err = mUdpcon_recv.doRecv( buf_recv, MAXPINGMSG );
        gettimeofday( &rcvtime, NULL );

        // keep track of failures and quit if have too many
        if( err <= 0 ){
      	    numberfail++;
            memset( (void*)&rcvtime, 0, sizeof(struct timeval) );
            printDebug(3, " 1) bouncePingTest - TIMEOUT bounce no message  numberfail=%d\n", numberfail);
        }
        else if( strcmp(buf_send, buf_recv) != 0 ){
	    printDebug(3,"3) Received Bounce message\n");

            // check that sender and receiver are correct
            msg = (char *)calloc( sizeof(buf_recv), sizeof(char) );
            strncpy( msg, buf_recv, sizeof(buf_recv) );
            serv = strtok( msg, " " );
            receiver = strtok( NULL, " " );

            // check whether this ping is from my server and intended for me
            if( strcmp(bPeeraddr.c_str(), serv )){
                printDebug(3, "*This is not my server : %s  msg: %s\n", bPeeraddr.c_str(), serv);
	    }
            else if( strcmp( bLocalName.c_str(), receiver )){
		printDebug(3,"*This is not for me: %s   msg: %s\n", bLocalName.c_str(), receiver);
	    }
	    else{
           	printDebug(3,"3) Received Bounced message  numberfail=%d\n", numberfail);
	        numberfail = 0;
		//ready_to_send = 1;
		send_sec = strtok( NULL, " " );
		send_usec = strtok( NULL, " " );
		prev = strtok( NULL, " " );
		printDebug(3, " 1) bounce received message from sendPingTest: %s\n", buf_recv );

                // convert the previous ping round trip time to a double and record it
                sscanf( prev, "%lf", &dblprev );

                if( dblprev >= 0.0 ){
                    mLastrtt[mNumbertests%AVGWINDOW] = dblprev;
                    ++mNumbertests;
          	}

                if( mNumbertests > mTestlimit )
               	    break;

            	 // fill in the ping packet
            	 memset( (void*)&buf_send, 0, sizeof(buf_send) );
            	 sprintf( buf_send, format1.c_str(), serv, receiver, send_sec, send_usec);

            	 // need to determine how much time we have been processing this ping packet
            	 memset( (void*)&sndtime, 0, sizeof(struct timeval) );
            	 gettimeofday( &sndtime, NULL );
            	 difftime = subtimev( &sndtime, &rcvtime );

            	// put the time we have spent processing the ping packet into the packet
            	// so the other end can subtract it off of the ping round trip time.
            	sprintf( buf2, format2.c_str(), difftime.tv_sec, difftime.tv_usec );
            	strcat( buf_send, buf2 );

        	// ********* send the ping back to the server ***************
		do{
            	     err = mUdpcon_send.doSend( buf_send, strlen( buf_send ));
            	    if( err<0 )
               	    	cerr << "ERROR on multicast bounce ping send\n";

		}while( err<0 && numberfail<10 );

                printDebug(3, " 2) Message sent by client: %s\n", buf_send );
            	free( msg );
        	printDebug(3, "  mTestlimit = %d      mNumbertests = %d \n", mTestlimit, mNumbertests );
	    }
        }
	else
       	    printDebug(3, " 3) bounce received own message: %s\n", buf_recv );
    } while( numberfail<10 && mNumbertests<mTestlimit );

    if( numberfail>10 ){
	for( int i=mNumbertests%AVGWINDOW; i>=0; i++ )
            mLastrtt[i] = -1.0;
    }
    return mNumbertests;
}


/* ---------------------------------------------------------
 *  calculateRTT - called by the ping server side to calculate
 *  the round trip time after a ping packet has bounced back.
 *  arguments
 *        pingmsg - the ping packet as it was received
 *        rcvtime - the time when this packet was received
 * ---------------------------------------------------------*/

double MULTIPing::calculateRTT( char *pingmsg,
			      struct timeval *rcvtime)
{
    int    index = 0;
    struct timeval sndtime;
    struct timeval delay;
    char   sender[MAXHOSTNAMELEN];
    char   receiver[MAXHOSTNAMELEN];
    struct timeval rtt;
    
    sscanf( pingmsg, "%s %s %ld %ld %ld %ld", sender, receiver,
	    &(sndtime.tv_sec), &(sndtime.tv_usec), &(delay.tv_sec),
	    &(delay.tv_usec) );

    rtt = subtimev( &(subtimev( rcvtime, &sndtime )), &delay );

    index = mNumbertests%AVGWINDOW;
    mLastrtt[index] = timevtomsec(&rtt);
    return mLastrtt[index];
}


/* ---------------------------------------------------------
 *  ThreadMain - this is the routine that is called when the
 *  thread is run using pthread_create.  It launches the ping
 *  test by calling the appropriate routine.
 * ---------------------------------------------------------*/
 
void MULTIPing::ThreadMain( void )
{
    int status;
    printDebug(3, "Starting ping test\n");

    if( mPinginf.isSender() )
	status = sendPingTest( );
    else
        status = bouncePingTest( );

    if( status < 0 )
	cerr << "ERROR running udpping" << endl;
    else
	printDebug(3, "Done with ping test\n");

    return;
}


/* ---------------------------------------------------------
 *  printStats - this routine prints the current test
 *  results.  It can be called asynchronously and it is
 *  also called when the ping test finishes.  This routine
 *  is thread safe (a mutex locks the statistics when they
 *  are being updated)
 * ---------------------------------------------------------*/

double MULTIPing::printStats()
{
    int    index = 0;
    int i = 0;
    double maxrtt, minrtt, avgrtt, curlast;
    
    index = (mNumbertests%AVGWINDOW) - 1;

    avgrtt = 0.0;
    for( i=0; i< AVGWINDOW && i < mNumbertests ; i++ ){
	if( i == 0 ){
	    maxrtt = mLastrtt[i];
	    minrtt = mLastrtt[i];
	}
	else{
	    if( maxrtt < mLastrtt[i] )
	        maxrtt = mLastrtt[i];
	    if( minrtt > mLastrtt[i] )
	        minrtt = mLastrtt[i];
	}
	avgrtt += mLastrtt[i];
    }
    curlast = mLastrtt[index];

    // remove the high value and the low value from the average
    if( i > 4 ){
	avgrtt = avgrtt - minrtt - maxrtt;
	i -= 2;
    }
    avgrtt = avgrtt / ((double)i);

    return curlast;
}


/* ---------------------------------------------------------
 *  getAverageRTT 
 *
 * ---------------------------------------------------------*/

double MULTIPing::getAverageRTT()
{
    double avgrtt = 0.0;

    int i;
    for( i=0; i< AVGWINDOW && i < mNumbertests ; i++ )
    	avgrtt += mLastrtt[i];

    avgrtt = avgrtt / ((double)i);

    return avgrtt;
}


/* ---------------------------------------------------------
 *  printResults:
 *
 * ---------------------------------------------------------*/

void
MULTIPing::printResults( void )
{
    PingResults pingResults( mPinginf );
    pingResults.fillStats( mLastrtt, mNumbertests );
    //(mPinginf.getResults())->newPingResults( &mPinginf, mStats_mutex, mLastrtt, mNumbertests );
    (mPinginf.getResults())->newPingResults( pingResults );
}


/* ---------------------------------------------------------
 *  printVariables:
 *
 * ---------------------------------------------------------*/

void
MULTIPing::printVariables( void )
{
    cout << "***MULTIPing:: printVariables\n";

    cout << "Local Name          = " << mPinginf.getLocalName() << endl; // the address of this machine
    cout << "Multicast Address   = " << mPinginf.getMcastAddr() << endl; // multicast address
    cout << "Test Peer address   = " << mPinginf.getRemoteTestPeer() << endl;  // test peer address

    cout << "mDelaypings=" << mDelaypings << endl; // how many msecs to delay between each ping test
                                                   // minimum delay will be the round trip time

    cout << "isSender     =" << mPinginf.isSender() << endl; // is this the server for the test
    cout << "mNumbertests=" << mNumbertests << endl;         // how many ping tests have been run so far
    cout << "mTestlimit  =" << mTestlimit << endl;           // number of ping tests to run
    cout << "PingTestInfo mPinginf\n";

    mPinginf.printVariables( cout );   // the properties for this ping test

    cout << "***\n";
}
